@@ -1,18 +1,28 @@
 from functools import partial
 
-from django.contrib.admin.util import flatten_fieldsets
+from django.contrib.admin.utils import flatten_fieldsets
 from django.contrib.admin.options import ModelAdmin, InlineModelAdmin
+from django import forms
+from django.forms.models import modelform_defines_fields
+
+from app_data.forms import multiform_factory, multiinlineformset_factory, MultiForm, AppDataBaseInlineFormSet
 
-from app_data.forms import multiform_factory, multiinlineformset_factory, MultiForm
 
 class AppDataAdminMixin(object):
 
     multiform = MultiForm
     app_form_opts = {}
 
+    def get_fieldsets(self, request, obj=None):
+        try:
+            return self.declared_fieldsets
+        except AttributeError:
+            return super(AppDataAdminMixin, self).get_fieldsets(request, obj)
+
     def _get_form_factory_opts(self, request, obj=None, **kwargs):
-        if self.declared_fieldsets:
-            fields = flatten_fieldsets(self.declared_fieldsets)
+        fieldsets = self.get_fieldsets(request, obj)
+        if fieldsets:
+            fields = flatten_fieldsets(fieldsets)
         else:
             fields = None
         if self.exclude is None:
@@ -61,11 +71,17 @@ class AppDataAdminMixin(object):
             "formfield_callback": partial(self.formfield_for_dbfield, request=request),
         }
         defaults.update(kwargs)
+
+        if hasattr(forms, 'ALL_FIELDS'):
+            # Django 1.7
+            if defaults['fields'] is None and not modelform_defines_fields(defaults['form']):
+                defaults['fields'] = forms.ALL_FIELDS
+
         return defaults
 
 
 class AppDataModelAdmin(AppDataAdminMixin, ModelAdmin):
-    def get_form(self, request, obj=None, **kwargs):
+    def get_form(self, request, obj=None, change=False, **kwargs):
         """
         Returns a Form class for use in the admin add view. This is used by
         add_view and change_view.
@@ -74,10 +90,13 @@ class AppDataModelAdmin(AppDataAdminMixin, ModelAdmin):
             return super(AppDataModelAdmin, self).get_form(request, obj=obj, **kwargs)
         return multiform_factory(self.model, **self._get_form_factory_opts(request, obj, **kwargs))
 
+
 class AppDataInlineModelAdmin(AppDataAdminMixin, InlineModelAdmin):
+    formset = AppDataBaseInlineFormSet
+
     def get_formset(self, request, obj=None, **kwargs):
         if self.multiform is None:
-            return super(AppDataModelAdmin, self).get_formset(request, obj=obj, **kwargs)
+            return super(AppDataInlineModelAdmin, self).get_formset(request, obj=obj, **kwargs)
         can_delete = self.can_delete
         if hasattr(self, 'has_delete_permission'):
             can_delete = can_delete and self.has_delete_permission(request, obj)
@@ -89,12 +108,13 @@ class AppDataInlineModelAdmin(AppDataAdminMixin, InlineModelAdmin):
             "can_delete": can_delete,
         }
         defaults.update(self._get_form_factory_opts(request, obj, **kwargs))
-
         return multiinlineformset_factory(self.parent_model, self.model, **defaults)
 
+
 class AppDataStackedInline(AppDataInlineModelAdmin):
     template = 'admin/edit_inline/stacked.html'
 
+
 class AppDataTabularInline(AppDataInlineModelAdmin):
     template = 'admin/edit_inline/tabular.html'
 
