@@ -1,20 +1,22 @@
-try:
-    from operator import methodcaller
-except ImportError:
-    methodcaller = lambda name: lambda o: getattr(o, name)()
+from operator import methodcaller
+from copy import deepcopy
+
+import six
 
 from django.forms.forms import NON_FIELD_ERRORS, Form
 from django.forms.formsets import formset_factory
 from django.forms.models import modelform_factory, _get_foreign_key, BaseInlineFormSet, BaseModelFormSet
+from django.forms.utils import pretty_name
 from django.utils.safestring import mark_safe
 
+
 class AppDataForm(Form):
-    def __init__(self, app_container, data=None, files=None, fields=(), exclude=(), **kwargs):
+    def __init__(self, app_container, data=None, files=None, fields=(), exclude=(), *args, **kwargs):
         self.app_container = app_container
-        super(AppDataForm, self).__init__(data, files, **kwargs)
+        super(AppDataForm, self).__init__(data, files, *args, **kwargs)
 
         if fields or exclude:
-            for f in self.fields.keys():
+            for f in list(self.fields.keys()):
                 if fields and f not in fields:
                     del self.fields[f]
                 elif f in exclude:
@@ -27,8 +29,9 @@ class AppDataForm(Form):
     def save(self):
         self.app_container.update(self.cleaned_data)
 
+
 class BaseFieldsDescriptor(object):
-    " Combines the base_fiels and prefixes them properly. Descriptor because needed on class level. "
+    """Combines the base_fields and prefixes them properly. Descriptor because needed on class level."""
     def __get__(self, instance, owner):
 
         if not hasattr(self, '_base_fields'):
@@ -41,11 +44,11 @@ class BaseFieldsDescriptor(object):
             bf.update(owner.ModelForm.base_fields)
 
             # go through all the app forms...
-            for label, opts in owner.get_app_form_opts().iteritems():
+            for label, opts in six.iteritems(owner.get_app_form_opts()):
                 Form = app_container[label].form_class
                 exclude = set(opts.get('exclude', ()))
                 fields = opts.get('fields', None)
-                for name, field in Form.base_fields.iteritems():
+                for name, field in six.iteritems(Form.base_fields):
                     # skip proper fields
                     if fields is not None and name not in fields:
                         continue
@@ -56,6 +59,7 @@ class BaseFieldsDescriptor(object):
 
         return self._base_fields
 
+
 class AppFormOptsDescriptor(object):
     def __get__(self, instance, owner):
         # we cannot check hasattr because parent's app_form_opts would pick it up
@@ -63,13 +67,26 @@ class AppFormOptsDescriptor(object):
             setattr(owner, '_app_form_opts', {})
         return owner._app_form_opts
 
-class MultiForm(object):
+
+class MultiFormMetaclass(type):
+    # This property is needed by BaseInlineFormSet which expect the form *class* to have a "real" _meta
+    # and thus the proxing in the instance property won't work
+    @property
+    def _meta(cls):
+        return cls.ModelForm._meta
+
+
+class MultiForm(six.with_metaclass(MultiFormMetaclass, object)):
     app_data_field = 'app_data'
     app_form_opts = AppFormOptsDescriptor()
 
     def __init__(self, *args, **kwargs):
         # construct the main model form
         self.model_form = self.ModelForm(*args, **kwargs)
+        try:
+            self.label_suffix = self.model_form.label_suffix
+        except AttributeError:
+            pass
         if self.model_form.is_bound:
             data = self.model_form.data
             files = self.model_form.files
@@ -79,7 +96,7 @@ class MultiForm(object):
         # construct all the app forms
         self.app_forms = {}
         app_container = getattr(self.model_form.instance, self.app_data_field)
-        for label, label_opts in self.get_app_form_opts().iteritems():
+        for label, label_opts in six.iteritems(self.get_app_form_opts()):
             prefix = label
             if self.model_form.prefix:
                 prefix = '%s-%s' % (self.model_form.prefix, prefix)
@@ -87,7 +104,7 @@ class MultiForm(object):
 
     @classmethod
     def get_app_form_opts(cls):
-        " Utility method to combinte app_form_opts from all base classes. "
+        """Utility method to combine app_form_opts from all base classes."""
         # subclass may wish to remove superclass's app_form
         skip_labels = set()
 
@@ -97,7 +114,7 @@ class MultiForm(object):
             # not a MultiForm, skip
             if not hasattr(c, 'app_form_opts'):
                 continue
-            for label, label_opts in c.app_form_opts.iteritems():
+            for label, label_opts in six.iteritems(c.app_form_opts):
                 if label in form_opts or label in skip_labels:
                     # form already defined, or should be skipped
                     continue
@@ -113,13 +130,13 @@ class MultiForm(object):
 
     @classmethod
     def add_form(cls, label, form_options={}):
-        " Add an app_data form to the multi form after it's creation. "
+        """Add an app_data form to the multi form after its creation."""
         cls.app_form_opts[label] = form_options.copy()
 
     @classmethod
     def remove_form(cls, label):
         """
-        Remove an app_data form to the multi form after it's creation.
+        Remove an app_data form to the multi form after its creation.
         Even if this form would be specified in a superclass it would be skipped.
         """
         cls.app_form_opts[label] = None
@@ -133,6 +150,10 @@ class MultiForm(object):
     def cleaned_data(self):
         return self.model_form.cleaned_data
 
+    @property
+    def is_multipart(self):
+        return self.model_form.is_multipart
+
     @property
     def _meta(self):
         # user by BaseInlineFormSet.add_fields
@@ -169,7 +190,7 @@ class MultiForm(object):
 
     def _get_all_forms(self):
         yield self.model_form
-        for f in self.app_forms.itervalues():
+        for f in six.itervalues(self.app_forms):
             yield f
 
     def __unicode__(self):
@@ -215,7 +236,7 @@ class MultiForm(object):
     def changed_data(self):
         if not hasattr(self, '_changed_data'):
             self._changed_data = cd = self.model_form.changed_data[:]
-            for label, form in self.app_forms.iteritems():
+            for label, form in six.iteritems(self.app_forms):
                 cd.extend(map(lambda n: '%s.%s' % (label, n), form.changed_data))
         return self._changed_data
 
@@ -224,8 +245,8 @@ class MultiForm(object):
         # combine all the errors
         if not hasattr(self, '_errors'):
             self._errors = self.model_form.errors.copy()
-            for label, form in self.app_forms.iteritems():
-                for k, v in form.errors.iteritems():
+            for label, form in six.iteritems(self.app_forms):
+                for k, v in six.iteritems(form.errors):
                     if k == NON_FIELD_ERRORS:
                         self._errors.setdefault(k, self.model_form.error_class()).extend(v)
                     else:
@@ -243,6 +264,18 @@ class MultiForm(object):
         return self.model_form.save(**kwargs)
 
 
+class AppDataBaseInlineFormSet(BaseInlineFormSet):
+
+    def add_fields(self, form, index):
+        """appcontainer fields are no longer added to the empty form, we can inject them hooking here."""
+        super(AppDataBaseInlineFormSet, self).add_fields(form, index)
+        for name, field in form.base_fields.items():
+            if name not in form.fields:
+                form.fields[name] = deepcopy(field)
+                if not form.fields[name].label:
+                    form.fields[name].label = pretty_name(name.split('.')[1])
+
+
 def multiform_factory(model, multiform=MultiForm, app_data_field='app_data', name=None, form_opts={}, **kwargs):
     model_form = modelform_factory(model, **kwargs)
     name = name or '%sWithAppDataForm' % model_form._meta.model.__name__
@@ -251,6 +284,7 @@ def multiform_factory(model, multiform=MultiForm, app_data_field='app_data', nam
         {'ModelForm': model_form, 'app_data_field': app_data_field, '_app_form_opts': form_opts}
     )
 
+
 def multiformset_factory(model, multiform=MultiForm, app_data_field='app_data', name=None, form_opts={},
                          formset=BaseModelFormSet, extra=3, can_order=False, can_delete=True, max_num=None,
                          **kwargs):
@@ -259,12 +293,12 @@ def multiformset_factory(model, multiform=MultiForm, app_data_field='app_data',
     FormSet.model = model
     return FormSet
 
+
 def multiinlineformset_factory(parent_model, model, multiform=MultiForm, app_data_field='app_data', name=None, form_opts={},
                                 formset=BaseInlineFormSet, fk_name=None, **kwargs):
     fk = _get_foreign_key(parent_model, model, fk_name=fk_name)
     if fk.unique:
         kwargs['max_num'] = 1
-
     FormSet = multiformset_factory(model, multiform, app_data_field, name, form_opts, formset=formset, **kwargs)
     FormSet.fk = fk
     return FormSet
